#!/usr/bin/perl -w
# GPCRHMM.pl script to run the GPCRHMM method
# Written by Lukas Kall, University of Washington
#
# Please cite 
#
# Markus Wistrand*, Lukas Kall* and Erik L.L. Sonnhammer.
# A general model of G protein-coupled receptor sequences and its 
# application to detect remote homologs.
# Protein Science, 15(3):509-21, Mars 2006.
# *These authors contributed equally to this work. 

my $usage = "gpcrhmm.pl [-h -c -g -l] input.fasta 
options: -h this helppage
         -c calc local score (default on)
         -g global score threshold (default = 0.0}
         -l local score threshold (default = 0.0)
";


use strict;
use IO::File;
use Getopt::Long;
use IPC::Open2;
use FindBin;
my $bindir =      "$FindBin::RealBin";
use Getopt::Long;


##################################
# Process input alternatives

my $doLocal = 1;
my $thrGlobal = 0;
my $thrLocal = 0;
my $help = 0;
# Process options
my $result = GetOptions(
                'help!' => \$help,
                'calc_local!' => \$doLocal,
                'global_threshold=f' => \$thrGlobal,
                'local_threshold=f' => \$thrLocal);
die ("Error on command line") unless $result;
die $usage if ($help);
die "No fasta file given as argument\n\n$usage" if (not defined $ARGV[0]);
my $inputFile = $ARGV[0];


my (
    $lthFilter,
    $decodeGlobal,
    $decodeLocal1,
    $decodeLocal2,
    $fileGlobal,
    $currentID,
    $currentScore,
    %hits,
    %scoreGlobal,
    %scoreLocal,
    $fastaCurrent,
    $runLocal,
    $currentLocal,
    $totLocal,
    $i,
    $j,
    $pid,
    $modelG,
    $modelL,
    $nullFile,
    @allIDs,	
  );



# Set parameters
$modelG        = "$bindir/model.bsmod";
$modelL        = "$bindir/model.trunc";
$nullFile      = "$bindir/dec.opt";
$lthFilter     = "$bindir/lengthFilter.awk";
$decodeGlobal  = "$lthFilter | $bindir/decodeanhmm -f $nullFile -ps $modelG ";
$decodeLocal1  = "$bindir/decodeanhmm -N 1 -pp -ps $modelG 2> /dev/null";
$decodeLocal2  = "$bindir/decodeanhmm -f $nullFile $modelL 2> /dev/null";
$fileGlobal    = new IO::File;

#####################################
# Reading all identifiers prior scoring the sequences
$fileGlobal -> open("<$inputFile") || die "could not open file";

while(<$fileGlobal>){
  if (/^>(\S+)/) { # New sequence ID
    push @allIDs, $1;
  }	
  elsif (/^[A-Za-z\n ]+$/) {}  
  else {
  m/([^A-Za-z ])/g; 
  die "The sequence $allIDs[$#allIDs] contains letter $1 which is not in the allowed alphabet!\n";}
}
close($fileGlobal);					
  	
#####################################
# Check if the seqs are long enough	
$fileGlobal -> open("cat  $inputFile | $lthFilter 2> /dev/null |");
if(length(<$fileGlobal>) eq 0) {die "All sequences are too short for GPCRHMM (<200 aa)\n";}
while(<$fileGlobal>) {;};
close($fileGlobal);
	
	
#####################################
# Reading from file or STDIN
#$fileGlobal -> open("$decodeGlobal  $inputFile 2>/dev/null | ") | die "could not run $decodeGlobal";
$fileGlobal -> open("cat $inputFile | $decodeGlobal 2>/dev/null | ") || die "could not run $decodeGlobal";

################################################################
$i        = 0;     # index for the number of sequences run
$j        = 0;     # index for local scoring
$totLocal = 0;     # total no of seqs to run in local
$runLocal = 0;     # indicating if there is a local sequence to be run
if ($doLocal) {
  printf "%-32s %-10s %-10s %-5s\n","Sequence identifier","global","local","pred","\n";
} else {
  printf "%-32s %-10s %-5s\n","Sequence identifier","global","pred","\n";
}
while(<$fileGlobal>)
  {
    if (/^>(\S+)/) { # New sequence ID	
      if($runLocal eq 1) {
	$scoreLocal{$currentID} = localScoring($fastaCurrent, $decodeLocal1, $decodeLocal2);
      }
      if ($i ne 0)    {printRun($currentID, $scoreGlobal{$currentID}, $scoreLocal{$currentID},$thrLocal); }      
      while (($1 ne $allIDs[$i]) && ($#allIDs>=$i)) {
        printToShort($allIDs[$i]);
	$i++;
      }       
      $runLocal = 0;
      $currentID = $1;
      $i++;
    }
    elsif (/^%score\sFW\s\S+\s\S+\s\S+\s(\S+)\s/) {  # Get the global score and save it
      $currentScore = $1; 
      $scoreGlobal{$currentID} = $1;
      
      if ( $currentScore > $thrGlobal) {    # Note whether this is a hit -> we must run local scoring
	$hits{$currentID} = 1; 
	$fastaCurrent = ">$currentID\n";
	$currentLocal = $currentID;
	$runLocal = 1;
	$totLocal++;
      }
      else { $hits{$currentID} = 0; $scoreLocal{$currentID} = "-";}
    }
    elsif (/^\s{3}([A-Z\s]+)\n/ && $hits{$currentID} eq 1) { 
      my $row = $1;
      $row =~ s/\s//g;
      $fastaCurrent .= $row;
    }
  }

# Running the last Local scoring
if($runLocal eq 1) {
  $scoreLocal{$currentID} = localScoring($fastaCurrent, $decodeLocal1, $decodeLocal2);
}
if ($i ne 0)    {printRun($currentID, $scoreGlobal{$currentID}, $scoreLocal{$currentID}, $thrLocal);}

# If the last sequence(s) is too short we must print so
while ($#allIDs>=$i) {
  printToShort($allIDs[$i]);
  $i++;
}
            
close($fileGlobal);

sub localScoring {
  return 0 if not $doLocal;
  
  my ($fileLocal,
      $output,
      $seqLocal,
      $label,
      $sequence,
      $firstM,
      $lastM,
      $start,
      $end,
      $localScore,
      $pid,	
     );
  my ($fasta, $decode1, $decode2) = @_;
  
  #run global part of local scoring
  $pid = open2(*GOUT, *FASTA, $decode1);
  $fasta = $fasta."\n";
  print FASTA $fasta;  
  close(FASTA);
  
  while(<GOUT>) {
    if (/^>(\S+)/) { # Sequence ID
    }
    elsif (/^\s{3}([A-Z\s]+)\n/) { # amino acid sequence
      my $row = $1;
      $row =~ s/\s//g;
      $sequence = $sequence.$row; 
    } 
    elsif (/^\?0\s([oOMncChi\s]+)\n/) { # sequence of labels
      my $row = $1;
      $row =~ s/\s//g;
      $label = $label.$row; 
    }
  }
  close(GOUT);
  waitpid($pid,0) if $pid;
  ($start, $end) = getLocalSeq($label);   # Indices to first and last amino acid labelled with an M
  $seqLocal = ">idLocal\n"; # necessary for decodeanhmm to work
  $seqLocal = $seqLocal.substr $sequence,$start, $end-$start+1;  # add the local sequence
  #run local part of local scoring
  $pid = open2(*GOUT2, *FASTA2, $decode2);
  print FASTA2 $seqLocal;  
  close(FASTA2);

LINE:  while(<GOUT2>) {
    if (/^>(\S+)/) {} # Sequence ID
    if (/%score FW \S+\s\S+\s\S+\s(\S+)\s/) {  # Get the local score
      $localScore = $1;
    }
  }
  close(GOUT2);
  waitpid($pid,0) if $pid;

if (defined($localScore)) {return $localScore;}
  else {return "na";}
}


sub getLocalSeq {
  my ($start,
      $end,
      $i,
      @aLabel,
     );
  my ($label) = @_;
  
  @aLabel = split("", $label);
  $i=0;
  
  # Find first M
 
  while($i<=$#aLabel) {
    if ( $aLabel[$i] =~ m/M/) {$start = $i; $i=$#aLabel;}
    $i++;
  }
  # Find last M
  $i = $#aLabel; # index of last element in array
  while($i>-2) {
    if ( $aLabel[$i] =~ m/M/) {$end = $i;$i=-2; }
    $i--;
  }
  return($start,$end);
}


sub printRun {
  my ($id, $Global,$Local, $thrL) = @_;
  my ($strLen, $strLenDefault, $hit);

  $hit = "No";
  $hit = "GPCR" if ($Local ne "-" && $Local >= $thrL);
  if (not $doLocal) {
    printf "%-30s %s %-10.2f %-5s\n", $id, " ", $Global, $hit,"\n";
  } elsif ($Local eq "-") {
    printf "%-30s %s %-10.2f %-10s %-5s\n", $id, " ", $Global, $Local, $hit,"\n";
  } else {  
    printf "%-30s %s %-10.2f %-10.2f %-5s\n", $id, " ", $Global, $Local, $hit, "\n"; 
  }	  		  
}


sub printToShort {
my ($id) = @_;
printf "%-30s %s %-21s %-5s\n", $id, " ","Too short sequence", "No";
}
